"""Low-level internal syncronous API"""

# We could have just made it so that one API handles both by returning awaitables when
# async stores are used, although that increaseas type unsafety and as of now does not play well
# with type checkers
from __future__ import annotations

from abc import abstractmethod
from operator import attrgetter
from time import monotonic as _now
from typing import (TYPE_CHECKING, Callable, Generic, Protocol, TypeVar, cast,
                    runtime_checkable)

from .errors import RateLimitError
from .rate import Rate, RateGroup
from .store import H, MemoryStore, T

__all__ = (
    "SyncStore",
    "SyncMemoryStore",
    "SyncRateLimit"
)

G = TypeVar("G")

if TYPE_CHECKING:
    from .rate import Rate, RateGroup

@runtime_checkable
class SyncStore(Protocol[T]):
    limit: SyncRateLimit

    def setup(self, ratelimit: SyncRateLimit):
        """Same as :meth:`uprate.store.BaseStore.setup`"""
        self.limit = ratelimit

    @abstractmethod
    def acquire(self, key: T) -> tuple[bool, float, Rate | None]:
        """Sync version of :meth:`uprate.store.BaseStore.acquire`"""
        ...

    @abstractmethod
    def reset(self, key: T) -> None:
        """Sync version of :meth:`uprate.store.BaseStore.reset`"""
        ...

    @abstractmethod
    def clear(self) -> None:
        """Sync version of :meth:`uprate.store.BaseStore.clear`"""
        ...

class SyncMemoryStore(SyncStore[H]):
    """An implementation of :class:`.SyncStore` protocol,
    hence is also the sync version of :class:`~uprate.store.MemoryStore`

    This implementation uses :class:`dict` and ejects stale buckets/keys
    periodically only when :meth:`.SyncMemoryStore.acquire` is called.

    This is a generic in TypeVar :data:`.H`

    Attributes
    ----------
    limit : :class:`uprate.ratelimit.RateLimit`
        The RateLimit to which this store is bound to.
    """
    _data: dict[H, tuple[list[int | float], ...]]

    def __init__(self):
        self._data = {}
        self._last_verified = 0.0
        self.limit = 0

    def setup(self, ratelimit: SyncRateLimit):
        super().setup(ratelimit)
        self._max_period = self.limit.rates[-1].period

    def acquire(self, key: H) -> tuple[bool, float, Rate | None]:
        now = _now()
        self.verify_cache() # Evict stale keys
        record = self._data.get(key, None)

        if record is None:
            # 1st insert
            self._data[key] = tuple([i.uses - 1, now] for i in self.limit.rates)
            return True, 0.0, None
        else:
            worst: float = False
            worst_rate: Rate | None = None

            for use_dt, rate in zip(record, self.limit.rates):
                if use_dt[0] == 0:
                    if (then := (use_dt[1] + rate.period)) <= now:
                        use_dt[:] = [rate.uses - 1, now]
                    elif (retry := then - now) > worst:
                        worst = retry
                        worst_rate = rate
                else:
                    use_dt[0] -= 1
            if worst is False:
                return True, 0.0, None

            return False, worst, worst_rate

    def reset(self, key: H) -> None:
        del self._data[key]

    def clear(self) -> None:
        self._data.clear()

    verify_cache = cast(Callable[[SyncStore[H]], None], MemoryStore.verify_cache)

class SyncRateLimit(Generic[G]):
    """Enforces multiple rates per provided keys.
    This is a low-level sync component.
    Sync version of :class:`uprate.ratelimit.RateLimit`

    Parameters
    ----------
    rate: :class:`~uprate.rate.Rate`, :class:`~uprate.rate.RateGroup`, (``Rate | RateGroup``)
        The rate(s) to enforce on keys in this ratelimits.

    store: :class:`.SyncStore`, (``SyncStore | None``)
        The sync store to use for this SyncRateLimit, If None then a :class:`.SyncMemoryStore` is used.
        By default, :data:`None`.

    Raises
    ------
    :exc:`TypeError`
        ``rate`` parameter provided to the constructor is of invalid type.

    Attributes
    ----------
    rates : tuple[:class:`uprate.rate.Rate`]
        A tuple of :class:`uprate.rate.Rate` sorted in ascending order by the rate's
        time period.
    store : :class:`.SyncStore`
        The sync store in use for this RateLimit.
    """
    rates: tuple[Rate, ...]
    store: SyncStore[G]

    def __init__(self, rate: Rate | RateGroup, store: SyncStore[G] | None = None) -> None:
        if isinstance(rate, Rate):
            self.rates = (rate,)
        elif isinstance(rate, RateGroup):
            rate._data.sort(key=attrgetter("period"))
            self.rates = tuple(rate._data)
        else:
            raise TypeError(f"Expected instance of uprate.rate.Rate or uprate.rate.RateGroup Instead got {type(rate)}")

        if store is None:
            self.store = SyncMemoryStore()
        elif isinstance(store, SyncStore):
            self.store = store
        else:
            raise TypeError("Expected a type deriving from uprate.store.SyncStore instead got " + str(type(store)))

        self.store.setup(self)

    def acquire(self, key: G) -> None:
        res, retry, rate = self.store.acquire(key)

        if not res:
            # <../ratelimit.py#82>
            raise RateLimitError(retry_after=retry, rate=rate) # type: ignore[arg-type]

    def reset(self, key: G = None) -> None:
        if not key:
            self.store.clear()
        else:
            self.store.reset(key)



"""Contains all Exception(s) uprate raises
"""

from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .rate import Rate

__all__ = (
    "RateLimitError",
)

class RateLimitError(Exception):
    """Raised when a Rate Limit is violated.

    Attributes
    ----------
    retry_after : :class:`float`
        The amount of time to retry after in seconds.
    rate : :class:`~uprate.rate.Rate`
        The rate that was violated
    """
    retry_after: float
    rate: Rate

    def __init__(self, retry_after: float, rate: Rate, *args):
        super().__init__(*args)
        self.retry_after = retry_after
        self.rate = rate

    def __float__(self) -> float:
        """Return :attr:`.RateLimitError.retry_after` as a float

        Returns
        -------
        float
            The amount in seconds to retry after.
        """
        return float(self.retry_after)

    def __int__(self) -> int:
        """Return :attr:`.RateLimitError.retry_after` as a ceiled int.

        Returns
        -------
        int
            The ceiled amount of seconds to retry after
        """
        return self.retry_after.__int__() + 1



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Generic, Literal, TypeVar

from .errors import RateLimitError
from .ratelimit import RateLimit

if TYPE_CHECKING:
    from .store import BaseStore
    from .rate import Rate, RateGroup

    from types import TracebackType

__all__ = (
    "Bucket",
)

T = TypeVar("T")

class BucketCM(Generic[T]):
    key: T
    bucket: Bucket[T]

    def __init__(self, bucket: Bucket[T], key: T):
        self.bucket = bucket
        self.key = key

    async def __aenter__(self):
        if self.bucket._queue:
            await self.bucket._queue.get()

        await self.__wait()
        return None

    async def __aexit__(self,
                        exc_type: type[BaseException] | None,
                        exc: BaseException | None,
                        tb: TracebackType | None) -> Literal[False]:
        if self.bucket._queue:
            self.bucket._queue.put_nowait(None)

        return False

    async def __wait(self) -> None:
        while True:
            try:
                await self.bucket._limit.acquire(self.key)
            except RateLimitError as err:
                await asyncio.sleep(float(err))
            else:
                return None

class Bucket(Generic[T]):
    """A high level ratelimit construct to obey both
    ratelimits and concurrency.

    .. note::

        Unlike other constructs in uprate, :class:`.Bucket` does not have
        a sync counterpart for threaded applications, a sync counterpart maybe
        added with enough interest.

    Parameters
    ----------
    rate : :class:`~uprate.rate.Rate`, :class:`~uprate.rate.RateGroup`, (``Rate | RateGroup``)
        The rate(s) to enforce on keys in this bucket.
    store : :class:`~uprate.store.BaseStore`, (``BaseStore | None``)
        The store to use for the under lying :class:`~uprate.ratelimit.RateLimit`. By default, :data:`None`.
    concurrency : :class:`int`
        The number of concurrently executing acquires/context managers.
        If ``0`` or lower then only the ratelimit is enforced.
        By default, ``0``
    """

    _limit: RateLimit[T]
    _queue: asyncio.Queue | None

    def __init__(self, rate: Rate | RateGroup, store: BaseStore[T] = None, concurrency: int = 0) -> None:
        self._limit = RateLimit(rate, store)

        if concurrency > 0:
            self._queue = asyncio.Queue(maxsize=concurrency)

            for _ in range(concurrency):
                self._queue.put_nowait(None)
        else:
            self._queue = None

    @classmethod
    def from_limit(cls, limit: RateLimit[T]) -> Bucket[T]:
        """Create a bucket from a :class:`~uprate.ratelimit.RateLimit`

        Parameters
        ----------
        limit : :class:`~uprate.ratelimit.RateLimit`
            The ratelimit to create a :class:`.Bucket` from.
            The provided ratelimit is used as it is, hence mutating it wil
            also affect the bucket

        Returns
        -------
        :class:`.Bucket`
            The created bucket
        """
        self = object.__new__(cls)
        self.limit = limit
        return self

    def acquire(self, key: T) -> BucketCM[T]:
        """Return an async context manager which
        tries to acquire a usage token upon entering while
        respecting the concurrency limit.

        Parameters
        ----------
        key : :data:`.T`
            The key to acquire.

        Returns
        -------
        :class:`uprate.bucket.BucketCM`
            The async context manager for the key.

        Example
        -------

        .. code-block:: python3

            bucket: Bucket[str] = Bucket(30 / Seconds(1), concurrency=1)
            ...
            # Wait until concurrency and ratelimit are satisfied.
            async with bucket.acquire("key"):
                ...
        """
        return BucketCM(self, key)

    async def reset(self, key: T = None) -> None:
        """Reset the given key.

        Parameters
        ----------
        key : :data:`.T`, :data:`None`, (``T | None``)
            The key to reset ratelimit for. If :data:`None`, then resets all ratelimits, by default :data:`None`.
        """
        self._limit.reset()

    @property
    def rates(self) -> tuple[Rate, ...]:
        """tuple[:class:`~uprate.rate.Rate`] : Same as :attr:`.RateLimit.rates`"""
        return self._limit.rates

    @property
    def store(self) -> BaseStore:
        """:class:`~uprate.store.BaseStore` : Same as :attr:`.RateLimit.store`"""
        return self._limit.store



from .rate import *
from .ratelimit import *
from .decorators import *
from .store import *
from ._sync import *
from .errors import *
from .bucket import *

from importlib.metadata import version

__version__ = version("uprate")



from __future__ import annotations

from asyncio import iscoroutinefunction, sleep
from collections.abc import Coroutine
from functools import wraps
from time import sleep as block
from typing import TYPE_CHECKING, Any, Callable, Protocol, TypeVar, Union, cast

from ._sync import SyncRateLimit, SyncStore
from ._utils import maybe_awaitable
from .errors import RateLimitError
from .ratelimit import RateLimit
from .store import BaseStore

if TYPE_CHECKING:
    from .rate import Rate, RateGroup

__all__ = (
    "on_retry_sleep",
    "on_retry_block",
    "ratelimit"
)

Key = TypeVar("Key")
"""An unbound and unconstrained TypeVar"""

R = TypeVar("R", covariant=True)

AnyRateLimit = Union[SyncRateLimit, RateLimit]

class LimitedCallable(Protocol[R]):
    limit: AnyRateLimit

    def __call__(self, *args, **kwds) -> R:
        ...

def _apply_attrs(func: Callable[..., R], **attrs) -> LimitedCallable[R]:
    func = cast(LimitedCallable[R], func)

    for k, v in attrs.items():
        setattr(func, k, v)

    return func

async def on_retry_sleep(error: RateLimitError) -> None:
    """Make the current task yield to the event_loop
    till a usage token is available for the rate limit
    which raised :exc:`uprate.errors.RateLimitError`

    Parameters
    ----------
    error : :exc:`.RateLimitError`
        The rate limit error to sleep for.
    """
    await sleep(error.retry_after)

def on_retry_block(error: RateLimitError) -> None:
    """Block the current thread till a usage token is
    available for the rate limit which raised
    :exc:`uprate.errors.RateLimitError`

    Parameters
    ----------
    error : :exc:`.RateLimitError`
        The rate limit error to block for.
    """
    block(error.retry_after)

# TODO: Complete Docs
def ratelimit(
    rate: Rate | RateGroup, *,
    key: Callable[..., Key] | Callable[..., Coroutine[Any, Any, Key]] | None = None,
    on_retry: Callable[[RateLimitError], Any] | Callable[[RateLimitError], Coroutine] | None = None,
    store: BaseStore | SyncStore | None = None
):
    """Limit a coroutine function or a callable to be called within
    provided rate. :ref:`Example here <index-example>`

    Parameters
    ----------
    rate: :class:`~uprate.rate.Rate`, :class:`~uprate.rate.RateGroup`, (``Rate | RateGroup``)
        The rate that the decorated function must follow.
    key : Callable[..., :data:`.Key`] | Callable[..., Coroutine[Any, Any, :data:`.Key`]] | :data:`None`
        The callback for generating a bucket for ratelimit from the arguments provided
        to the decorated function, this can be a coroutine function only when the decorated
        is a coroutine function as well. If :data:`None`, a default callback returning a string based on the
        decorated function's name is used, by default :data:`None`.
    on_retry : Callable[[:exc:`.RateLimitError`], Any] | Callable[[:exc:`.RateLimitError`], Coroutine] | :data:`None`
        If provided then this function will be called when the function gets ratelimited
        and then the decorated function will be called again, if :data:`None` then function call isn't retried
        and :exc:`.RateLimitError` is raised, by default :data:`None`.
    store : :class:`.BaseStore` | :class:`.SyncStore` | :data:`None`
        The store to use for the rate limit, must be of type :class:`.BaseStore` if
        decorated function is a coroutine function else :class:`.SyncStore`.
        If :data:`None` then a suitable derived memory store is used, by default :data:`None`.

    Raises
    ------
    :exc:`.RateLimitError`
        ``on_retry`` parameter is not provided and the decorated function got ratelimited.
    """
    def decorator(func: Callable[..., R]) -> LimitedCallable[R]:
        nonlocal on_retry, key
        key = key or cast(Callable[..., Key], lambda *a, **k: "DEFAULT_BUCKET_" + func.__name__)

        if iscoroutinefunction(func):
            if isinstance(store, BaseStore) or store is None:
                limit: AnyRateLimit = RateLimit(rate, store)
            else:
                raise TypeError("Cannot use a uprate._sync.SyncStore instance with a coroutine function.")

            @wraps(func)
            async def rated(*args, **kwargs):
                while True:
                    try:
                        bucket = await maybe_awaitable(key(*args, **kwargs))
                        await limit.acquire(bucket)
                    except RateLimitError as err:
                        if on_retry is None:
                            raise err from None
                        else:
                            await maybe_awaitable(on_retry(err))
                    else:
                        return await func(*args, **kwargs)
        else:
            if isinstance(store, SyncStore) or store is None:
                limit = SyncRateLimit(rate, store)
            else:
                raise TypeError("Cannot use a uprate.BaseStore instance with a subroutine.")

            @wraps(func)
            def rated(*args, **kwargs):
                while True:
                    try:
                        bucket = key(*args, **kwargs)
                        limit.acquire(bucket)
                    except RateLimitError as err:
                        if on_retry is None:
                            raise err from None
                        else:
                            on_retry(err)
                    else:
                        return func()

        return _apply_attrs(rated, limit=limit)

    return decorator




from __future__ import annotations

from abc import abstractmethod
from collections.abc import Hashable
from time import monotonic as _now
from typing import (TYPE_CHECKING, Optional, Protocol, TypeVar,
                    runtime_checkable)

if TYPE_CHECKING:
    from .rate import Rate
    from .ratelimit import RateLimit

__all__ = (
    "BaseStore",
    "MemoryStore"
)

T = TypeVar("T", contravariant=True)
"""An unbound and unconstrained contravariant TypeVar"""

H = TypeVar("H", contravariant=True, bound=Hashable)
"""A contravariant TypeVar bound to :class:`collections.abc.Hashable`"""

@runtime_checkable
class BaseStore(Protocol[T]):
    """A protocol defining the design of async stores.
    This is generic in TypeVar :data:`uprate.store.T` which is unbound and unconstrained.

    Subclasses implementing this protocol should inherit from this class.

    Attributes
    ----------
    limit : :class:`uprate.ratelimit.RateLimit`
        The RateLimit to which this store is bound to.
    """
    limit: RateLimit

    def setup(self, ratelimit: RateLimit):
        """Adds the ratelimit that this store is bound to as an
        attribute under :attr:`.BaseStore.limit`. This method exists
        only to create a circular reference between the store and ratelimit.
        :attr:`uprate.ratelimit.RateLimit.rates` attribute allows the store to access
        all implemented rates.

        Parameters
        ----------
        ratelimit : :class:`uprate.ratelimit.RateLimit`
            The ratelimit which this store is bound to.
        """
        self.limit = ratelimit

    @abstractmethod
    async def acquire(self, key: T) -> tuple[bool, float, Optional[Rate]]:
        """Try to acquire a usage token for given key.

        .. note::
            To get all the rates that this key follows use

            .. code-block:: python

                self.limit.rates

        .. note::
            If a HashMap like data-structure is being nested, then it's best that it is nested by
            the rates instead of the keys, since the number of keys may not exceed 1 in most cases,
            while the number of keys could grow upto 100k or more fairly quickly.

        Parameters
        ----------
        key : :data:`uprate.store.T`
            The key to acquire a ratelimit for.

        Returns
        -------
        tuple[:class:`bool`, :class:`float`, :class:`~uprate.rate.Rate` | :data:`None`]
            A three element tuple, the first element of type :class:`bool` depicting success.

            Second element :class:`float` which is the amount of time to retry in, If a usage
            token was acquired this should return ``0`` other-wise the time in which a
            usage token will be available. If the store does not support retry time then it
            should return a negative value like ``-1`` (negative values shall be returned only on
            failure if the retry time cannot be determined).

            The last element is the :class:`uprate.rate.Rate` object which was violated,
            this must be the rate which will take the longest to reset. The last element is
            expected to be :data:`None` if acquiring was successfull.
        """
        ...

    @abstractmethod
    async def reset(self, key: T) -> None:
        """Reset the usage tokens for given key.
        Implementation wise, deleting all the records for the given key should be enough.

        Parameters
        ----------
        key : :data:`uprate.store.T`
            The key to acquire a ratelimit for.
        """
        ...

    @abstractmethod
    async def clear(self) -> None:
        """Reset all the keys in the store.
        """
        ...

class MemoryStore(BaseStore[H]):
    """An implementation of :class:`.BaseStore` protocol.
    This implementation uses :class:`dict` and ejects stale buckets/keys
    periodically only when :meth:`.MemoryStore.acquire` is called.

    This is a generic in TypeVar :data:`.H`

    Attributes
    ----------
    limit : :class:`uprate.ratelimit.RateLimit`
        The RateLimit to which this store is bound to.
    """
    _data: dict[H, tuple[list[int | float], ...]]

    def __init__(self):
        self._data = {}
        self._last_verified = 0.0

    def setup(self, ratelimit: RateLimit):
        super().setup(ratelimit)
        self._max_period = self.limit.rates[-1].period

    async def acquire(self, key: H) -> tuple[bool, float, Optional[Rate]]:
        now = _now()
        # Would using loop.call_at be a better idea?
        # or per key scheduled callback maybe?
        self.verify_cache() # Evict stale keys
        record = self._data.get(key, None)

        if record is None:
            # 1st insert
            self._data[key] = tuple([i.uses - 1, now] for i in self.limit.rates)
            return True, 0.0, None
        else:
            worst: float = False
            worst_rate: Optional[Rate] = None

            # Optimisations: We do not need to update every rate
            # that expires only the ones that don't have usage tokens.
            for use_dt, rate in zip(record, self.limit.rates):
                if use_dt[0] == 0:
                    if (then := (use_dt[1] + rate.period)) <= now:
                        # We have no tokens left but the rate has expired
                        # so we reset it and acquire a token.
                        use_dt[:] = [rate.uses - 1, now]
                    elif (retry := then - now) > worst:
                        # no tokens and the rate has time left to expire.
                        worst = retry
                        worst_rate = rate
                else:
                    use_dt[0] -= 1
            if worst is False:
                return True, 0.0, None

            return False, worst, worst_rate

    async def reset(self, key: H) -> None:
        del self._data[key]

    async def clear(self) -> None:
        self._data.clear()

    def verify_cache(self) -> None:
        # There is no way something has expired since the last
        # check if enough time hasn't passed.
        now = _now()
        if (now - self._last_verified) < self._max_period:
            return

        delete = list[H]()

        for k, v in self._data.items():
            if self._max_period < (now - v[-1][1]):
                delete.append(k)

        for i in delete:
            del self._data[i]

        self._last_verified = _now()



from __future__ import annotations

from operator import attrgetter
from typing import Generic, Optional, TypeVar, Union

from uprate.store import BaseStore, MemoryStore

from .errors import RateLimitError
from .rate import Rate, RateGroup

__all__ = (
    "RateLimit",
)

H = TypeVar("H")
"""A TypeVar"""

class RateLimit(Generic[H]):
    """Enforces multiple rates per provided keys.
    This is a low-level component.

    Parameters
    ----------
    rate: :class:`~uprate.rate.Rate`, :class:`~uprate.rate.RateGroup`, (``Rate | RateGroup``)
        The rate(s) to enforce on keys in this ratelimits.

    store: :class:`~uprate.store.BaseStore`, (``BaseStore | None``)
        The store to use for this RateLimit, If None then a :class:`uprate.store.MemoryStore` is used.
        By default, :data:`None`.

    Raises
    ------
    :exc:`TypeError`
        ``rate`` parameter provided to the constructor is of invalid type.

    Attributes
    ----------
    rates : tuple[:class:`~uprate.rate.Rate`]
        A tuple of :class:`~uprate.rate.Rate` sorted in ascending order by the rate's
        time period.
    store : :class:`~uprate.store.BaseStore`
        The store in use for this RateLimit.
    """
    rates: tuple[Rate, ...]
    store: BaseStore[H]

    def __init__(self, rate: Union[Rate, RateGroup], store: Optional[BaseStore[H]] = None) -> None:
        if isinstance(rate, Rate):
            self.rates = (rate,)
        elif isinstance(rate, RateGroup):
            rate._data.sort(key=attrgetter("period"))
            self.rates = tuple(rate._data)
        else:
            raise TypeError(f"Expected instance of uprate.rate.Rate or uprate.rate.RateGroup Instead got {type(rate)}")

        if store is None:
            self.store = MemoryStore()
        elif isinstance(store, BaseStore):
            self.store = store
        else:
            raise TypeError("Expected a type deriving from uprate.store.BaseStore instead got " + str(type(store)))

        self.store.setup(self)

    async def acquire(self, key: H) -> None:
        """Try to acquire a usage token for given token.
        Raise an error if token can't be acquired.

        Parameters
        ----------
        key : :data:`.H`
            The key to acquire a usage token for.

        Raises
        ------
        :exc:`~uprate.errors.RateLimitError`
            Cannot acquire usage token due to exhaustion.
        """
        res, retry, rate = await self.store.acquire(key)

        if not res:
            # cast is ugly, overloads don't work (parameters don't change)
            # TODO: https://www.python.org/dev/peps/pep-0647/
            raise RateLimitError(retry_after=retry, rate=rate) # type: ignore[arg-type]

    async def reset(self, key: H = None) -> None:
        """Reset the given key.

        Parameters
        ----------
        key : :data:`.H`, :data:`None`, (``H | None``)
            The key to reset ratelimit for. If :data:`None`, then resets all ratelimits, by default :data:`None`.
        """
        if not key:
            await self.store.clear()
        else:
            await self.store.reset(key)



from __future__ import annotations

from inspect import isawaitable
from typing import Awaitable, TypeVar, Union

R = TypeVar("R")

__all__ = (
    "maybe_awaitable",
)

async def maybe_awaitable(ret: Union[R, Awaitable[R]]) -> R:
    if isawaitable(ret): # Type Guard Problem here
        return await ret # type: ignore[misc]
    else:
        return ret # type: ignore[return-value]



from __future__ import annotations

from typing import Union

__all__ = (
    "Rate",
    "Seconds",
    "Minutes",
    "Hours",
    "Days",
    "Weeks",
    "Months"
)

class RateGroup:
    __slots__ = ("_data",)

    _data: list[Rate]

    def __init__(self):
        self._data = []

    def __or__(self, other) -> RateGroup:
        if isinstance(other, Rate):
            self._data.append(other)
            return self
        elif isinstance(other, self.__class__):
            self._data = self._data + other._data
            return self
        return NotImplemented

    # __ror__ wouldn't be called
    # unless someone is using it wrong or
    # is dealing with uprate's internals
    __ror__ = __or__


class Rate:
    """A rate consisting of uses and time period

    Used to define rates for ratelimits
    with syntatical expressions.

    You should refrain from creating your own Rate objects
    and instead use the objects provided in the library.

    Attributes
    ----------
    uses: int
        Number of uses per time period
    period: float
        The time period of the rate in seconds
    """
    __slots__ = ("uses", "period")

    uses: int
    period: float

    def __init__(self, uses: int, period: float) -> None:
        self.uses = uses
        self.period = period

    def __call__(self, magnitude: float) -> Rate:
        return self.__class__(self.uses, self.period * magnitude)

    def __or__(self, other: Union[Rate, RateGroup]) -> RateGroup:
        if isinstance(other, RateGroup):
            return other | self
        elif isinstance(other, self.__class__):
            return RateGroup() | self | other
        return NotImplemented

    __ror__ = __or__

    def __rtruediv__(self, other: int) -> Rate:
        return self.__class__(other, self.period)

    def __mul__(self, other: float) -> Rate:
        # Let multipling a Rate increase the time period
        if isinstance(other, (float, int)):
            return self.__class__(self.uses, self.period * other)
        return NotImplemented

    __rmul__ = __mul__

    def __truediv__(self, other: float) -> Rate:
        # Let dividing a Rate decrease the time period
        if isinstance(other, (float, int)):
            return self.__class__(self.uses, self.period / other)
        return NotImplemented

    def __add__(self, other: Rate) -> Rate:
        if isinstance(other, self.__class__):
            if other.uses != 1 or self.uses != 1:
                raise ValueError("Cannot 'add' two rates which have uses other than one")
            return self.__class__(1, self.period + other.period)
        return NotImplemented

    __radd__ = __add__

    # object has other as object

    def __eq__(self, other) -> bool:
        if isinstance(other, self.__class__):
            return self.period == other.period and self.uses == other.uses
        return NotImplemented

    def __ne__(self, other) -> bool:
        if isinstance(other, self.__class__):
            return not (self == other)
        return False

    def __hash__(self) -> int:
        # This might be a bad idea
        return hash((self.uses, self.period))

    def __str__(self) -> str:
        return f"{self.uses}/{self.period}"


Seconds = Rate(1, 1)
"""A Rate of 1 use / 1 Second
Used in syntatical expression to define rate

Example
-------
.. code-block:: python

    assert 2 / Seconds(3) == 2 / (3 * Seconds)
    ...
    # 30 uses per 2 min 30 sec but also 2 uses per 2 seconds
    api_rate: uprate.rate.RateGroup = 2/Seconds(2) | 30/(2 * Minutes + 30 * Seconds)

"""

Minutes = 60 * Seconds
"""Rate of 1 use / 1 Minute"""

Hours = 60 * Minutes
"""Rate of 1 use / 1 Hour"""

Days = 24 * Hours
"""Rate of 1 use / 1 Day"""

Weeks = 7 * Days
"""Rate of 1 use / 1 Week"""

Months = 30 * Days
"""Rate of 1 use / 1 Month


Note
----
If you need larger time periods you can create them by

.. code-block:: python

    Year = Rate(1, 60 * 60 * 24 * 365)
    # Proceed to use like other unitary use rates
    rate = 5 / Year(2) # 5 uses per 2 years

"""



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['uprate']

package_data = \
{'': ['*']}

extras_require = \
{'docs': ['Sphinx>=4.1.1,<5.0.0',
          'myst-parser>=0.15.1,<0.16.0',
          'furo>=2021.7.5-beta.38,<2022.0.0',
          'sphinx-copybutton>=0.4.0,<0.5.0']}

setup_kwargs = {
    'name': 'uprate',
    'version': '0.2.1',
    'description': 'Ratelimits. Dumbified. A fully typed, simple ratelimit library.',
    'long_description': '<h1 align="center">\n    <img src="https://github.com/WizzyGeek/WizzyGeek/raw/main/assets/uprate/uprate_logo_rev2.png">\n</h1>\n\n<div align="center">\n    A fully typed, simple ratelimit library.\n</div>\n\n<div align="center">\n    <br/>\n    <img src="https://forthebadge.com/images/badges/made-with-python.svg">\n    <img src="https://forthebadge.com/images/badges/built-with-love.svg">\n</div>\n\n<hr/>\n\n# About\n\nUprate is a robust, simple ratelimit library.<br/>\nWhile providing a simple to use api, it is also highly scalable\nand provides absolute control for fine-tuning.<br/> Hence, Uprate\ncan be used in all stages of your app from prototyping to production! 🚀\n<br/>\n[Here](#example) is a simple example.\n\n## Why?\n\nThere are two ways ratelimits are implemented in python for server-side\n - Make everything from scratch\n - Use a framework dependent ratelimit library.\n\nThe problem in the first way is obvious, it\'s harder, consumes more time.<br/>\nUsing a framework dependent ratelimit library is more feasible, but often\nthese libraries don\'t provide features like external stores, multiple ratelimits\nand manual key specification. While there are some awesome ratelimit libraries for\nframwork X, not everyone uses framework X 😕.\n\nRatelimits in client-sided coded also face similar problems. Often APIs enforce multiple\nratelimits. Making a ratelimiter from scratch for your API wrapper\nor a small scale data collector takes up valuable dev time, which is why uprate aims to also\nprovide tools for client-sided code.\n\n## [Documentation](https://uprate.readthedocs.io/en/latest/)\n\nThe documentation can be found at <https://uprate.readthedocs.io/en/latest/> <br/>\n\n# Getting Started\n\n## Installation\n\nYou can install the latest stable version from pypi by\n```\npip install uprate\n```\n*or* you can install the dev version from github\n```\npip install git+https://github.com/WizzyGeek/uprate.git@master#egg=uprate\n```\n## Usage\n\n```\nimport uprate\n```\n\nAnd you are good to go! 🤘\n\n## Example\n\nHere is a simple example that demonstrates Uprate\'s Awesomeness.\n\n```py\nimport uprate\n\n@uprate.ratelimit(1 / (uprate.Seconds(2) + uprate.Minutes(1)))\ndef limited():\n    ...\n\nlimited()\n\ntry:\n    # Within 62 seconds\n    limited()\nexcept uprate.RateLimitError:\n    print("called function too fast")\n```\n\nAnd **so much more**!\n\n<div align="center">\n    <h1></h1>\n    <h6>Thanks to <a href="https://github.com/someonetookmycode">@someonetookmycode</a> for the graphical assets</h6>\n    <h6>© WizzyGeek 2021</h6>\n</div>\n',
    'author': 'WizzyGeek',
    'author_email': 'ojasscoding@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/WizzyGeek/uprate#readme',
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
